#version 430

// depth maps
const int OMNI_SHADOW_MAPS = 3;
const int SHADOW_MAPS = 2;
const int SPOT_SHADOW_MAPS = 1;
uniform float farPlane;
uniform samplerCube omniShadowMap[OMNI_SHADOW_MAPS];
uniform sampler2D directionalShadowMap[SHADOW_MAPS];
uniform sampler2D spotShadowMap[SPOT_SHADOW_MAPS];

// PCF
const int PCF_TOTAL_SAMPLES = 9;
const int PCF_COUNT = 1;
const int PCF_OMNI_SAMPLES = 20;

in VertexData {
	vec3 positionWorld;
	vec3 normalWorld;
	vec2 uvs;
    mat3 tbn;
    vec4 frag_pos_lightspace[SHADOW_MAPS];
	vec4 frag_pos_lightspace_spot[SPOT_SHADOW_MAPS];
} vert;

layout(location = 0) out vec4 color;
layout(location = 1) out vec4 bloomColor;

uniform float brightness; // adjusts general scene brightness

// textures
uniform sampler2D albedoMap;
uniform sampler2D normalMap;
uniform sampler2D aoRoughMetalMap; // r = ao, g = roughness, b = metallic

// IBL
uniform samplerCube irradianceMap;
uniform samplerCube prefilteredMap;
uniform sampler2D brdfTex;

// camera
uniform vec3 cameraWorld;

// lights
const int POINT_NUM = 5;
const int DIR_NUM = 5;
const int SPOT_NUM = 5;

// array sizes
uniform int actualPointSize;
uniform int actualDirectionalSize;
uniform int actualSpotSize;

struct PointLight {
	vec3 color;
    float PADDING1;
	vec3 position;
    float PADDING2;
	vec3 attenuation;
    float fallOfValue;
    int shadowFBO;    
	int depthMap;
    int isActive; 
    int volumetric;
};

struct DirectionalLight {
	vec3 color;
    float intensity;
	vec3 direction;
    float PADDING2;
    vec3 position;
    float PADDING3;
    int shadowFBO;
	int depthMap;
	int isActive;
	int PADDING5;
};

struct SpotLight {
    vec3 color;
	float cutOff;
	vec3 direction;
	float outerCutOff;
	vec3 attenuation;
    float PADDING1;
	vec3 position;
	float PADDING2;
	int shadowFBO;
	int depthMap;
	int isActive;
    int volumetric;
};

layout(std140) uniform pointBlock {
    PointLight pointLights[POINT_NUM];
};

layout(std140) uniform directionalBlock {
    DirectionalLight directionalLights[DIR_NUM];
};

layout(std140) uniform spotBlock {
    SpotLight spotLights[SPOT_NUM];
};

const float PI = 3.14159265359;


float DistributionGGX(vec3 N, vec3 H, float roughness);
float GeometrySchlickGGX(float NdotV, float roughness);
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness);
vec3 fresnelSchlick(float cosTheta, vec3 F0);
vec3 fresnelSchlickRoughness(float cosTheta, vec3 F0, float roughness);

float shadowCalculationPointLight(int lightIndex, vec3 fragPos);
float shadowCalculationDirectionalLight(int lightIndex, vec3 fragPos);
float shadowCalculationSpotLight(int lightIndex, vec3 fragPos);

vec3 sampleOffsetDirections[20] = vec3[]
(
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
);

void main()
{
    vec3 albedo     = pow(texture(albedoMap, vert.uvs).rgb, vec3(2.2));
    float ao        = texture(aoRoughMetalMap, vert.uvs).r;
    float roughness = texture(aoRoughMetalMap, vert.uvs).g;
    float metallic  = texture(aoRoughMetalMap, vert.uvs).b;
    vec3 normal     = texture(normalMap, vert.uvs).rgb;
    normal  	    = normal * 2.0 - 1.0;
    normal  	    = normalize(vert.tbn * normal);

    // normal and view direction
    vec3 N = normal;
    vec3 V = normalize(cameraWorld - vert.positionWorld);

    // reflect
    vec3 R = reflect(-V, N);

    // fresnel
    vec3 F0 = vec3(0.04); // base reflectivity
    F0 = mix(F0, albedo, metallic);

    // reflectance equation
    vec3 Lo = vec3(0.0);    
    
    
    for(int i = 0; i < actualDirectionalSize; ++i) // for each directional lights
    {
        if(directionalLights[i].isActive == 1)
        {
            // calculate per-light radiance
            vec3 L = normalize(-directionalLights[i].direction); // light direction
            vec3 H = normalize(V + L); // halfway vector
            vec3 radiance     = directionalLights[i].color;

            // cook-torrance brdf
            float NDF = DistributionGGX(N, H, roughness);
            float G   = GeometrySmith(N, V, L, roughness);
            vec3 F    = fresnelSchlick(max(dot(H, V), 0.0), F0);
            
            vec3 kS = F;
            vec3 kD = vec3(1.0) - kS;
            kD *= 1.0 - metallic;
            
            vec3 numerator    = NDF * G * F; // zaehler
            float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0); // nenner
            vec3 specular     = numerator / max(denominator, 0.001);
                
            // add to outgoing radiance Lo
            float NdotL = max(dot(N, L), 0.0);
            Lo += (kD * albedo / PI + specular) * radiance * NdotL * (1.0 - shadowCalculationDirectionalLight(i, vert.positionWorld));
        }  
    }
    
    for(int i = 0; i < actualPointSize; ++i) // for each point light
    {
        if(pointLights[i].isActive == 1)
        {
            // calculate per-light radiance
            vec3 L = normalize(pointLights[i].position - vert.positionWorld); // light direction
            vec3 H = normalize(V + L); // halfway vector
            float distance    = length(pointLights[i].position - vert.positionWorld);
            float attenuation = 1.0 / (pointLights[i].attenuation.x + pointLights[i].attenuation.y * distance + pointLights[i].attenuation.z * pow(distance, 2));
            vec3 radiance     = pointLights[i].color * attenuation;
            
            // cook-torrance brdf
            float NDF = DistributionGGX(N, H, roughness);
            float G   = GeometrySmith(N, V, L, roughness);
            vec3 F    = fresnelSchlick(max(dot(H, V), 0.0), F0);

            vec3 kS = F;
            vec3 kD = vec3(1.0) - kS;
            kD *= 1.0 - metallic;
            
            vec3 numerator    = NDF * G * F; // zaehler
            float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0); // nenner
            vec3 specular     = numerator / max(denominator, 0.001);
                
            // add to outgoing radiance Lo
            float NdotL = max(dot(N, L), 0.0);
            Lo += (kD * albedo / PI + specular) * radiance * NdotL * (1.0 - shadowCalculationPointLight(i, vert.positionWorld)); 
        } 
    }

    for(int i = 0; i < actualSpotSize; ++i)
    {
        if(spotLights[i].isActive == 1)
        {
            vec3 L = normalize(spotLights[i].position - vert.positionWorld); // light direction
            float theta = dot(L, normalize(-spotLights[i].direction));
            float epsilon   = spotLights[i].cutOff - spotLights[i].outerCutOff;
            float intensity = clamp((theta - (spotLights[i].outerCutOff)) / epsilon, 0.0, 1.0);

            if(intensity > 0)
            {
                vec3 H = normalize(V + L); // halfway vector
                float distance    = length(spotLights[i].position - vert.positionWorld);
                float attenuation = 1.0 / (spotLights[i].attenuation.x + spotLights[i].attenuation.y * distance + spotLights[i].attenuation.z * pow(distance, 2));
                vec3 radiance     = spotLights[i].color * attenuation;
                
                // cook-torrance brdf
                float NDF = DistributionGGX(N, H, roughness);
                float G   = GeometrySmith(N, V, L, roughness);
                vec3 F    = fresnelSchlick(max(dot(H, V), 0.0), F0);

                vec3 kS = F;
                vec3 kD = vec3(1.0) - kS;
                kD *= 1.0 - metallic;
                
                vec3 numerator    = NDF * G * F; // zaehler
                float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0); // nenner
                vec3 specular     = numerator / max(denominator, 0.001);
                    
                // add to outgoing radiance Lo
                float NdotL = max(dot(N, L), 0.0);
                Lo += (kD * albedo * intensity / PI + specular * intensity) * radiance * NdotL * (1.0 - shadowCalculationSpotLight(i, vert.positionWorld));
            }
        }
    }

    float NdotV = max(dot(N, V), 0.0);
    vec3 F = fresnelSchlickRoughness(NdotV, F0, roughness);
    vec3 kD = (1.0 - F) * (1.0 - metallic);
    vec3 irradiance = texture(irradianceMap, N).rgb;
    vec3 diffuse    = irradiance * albedo;      

    const float MAX_MIP_LEVEL = 4.0;
    vec3 prefilteredColor = textureLod(prefilteredMap, R,  roughness * MAX_MIP_LEVEL).rgb;    
    vec2 brdf  = texture(brdfTex, vec2(max(dot(N, V), 0.0), roughness)).rg;
    vec3 specular = prefilteredColor * (F * brdf.x + brdf.y);
    vec3 ambient = (kD * diffuse + specular) * ao;          
    vec3 col = ambient + Lo;
 
    col = col / (col + vec3(1.0));
    col = pow(col, vec3(1.0/2.2)); // gamma correction

    color = vec4(col, 1.0);

    float bright = dot(color.rgb, vec3(0.2126, 0.7152, 0.0722)); 
    color.rgb += brightness;

    //color.rgb = fract(vert.positionWorld);

    if(bright > 1.0)
        bloomColor = vec4(color.rgb, 1.0);
    else
        bloomColor = vec4(0.0, 0.0, 0.0, 1.0);

}

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;
	
    float num   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;
	
    return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float num   = NdotV;
    float denom = NdotV * (1.0 - k) + k;
	
    return num / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}

vec3 fresnelSchlickRoughness(float cosTheta, vec3 F0, float roughness)
{
    return F0 + (max(vec3(1.0 - roughness), F0) - F0) * pow(1.0 - cosTheta, 5.0);
}

float shadowCalculationPointLight(int lightIndex, vec3 fragPos)
{    
    vec3 fragToLight = fragPos - pointLights[lightIndex].position;    
    float currentDepth = length(fragToLight);

    /*float bias = 0.05; 
    float closestDepth = texture(omniShadowMap[lightIndex], fragToLight).r;
    closestDepth *= farPlane;
    float shadow = currentDepth -  bias > closestDepth ? 1.0 : 0.0;
    return shadow;*/

    float bias   = 0.0005;
    float shadow = 0.0;
	float view_distance = length(cameraWorld - fragPos);
	float disk_radius = (1.0 + (view_distance / farPlane)) / 120.0;
    
    for(int i = 0; i < PCF_OMNI_SAMPLES; ++i) {
	    float closestDepth = texture(omniShadowMap[lightIndex], (fragToLight + sampleOffsetDirections[i] * disk_radius)).r;    
        closestDepth *= farPlane;

		if(currentDepth - bias > closestDepth) {
			shadow += 1.0;
		}
	}
	return shadow / float(PCF_OMNI_SAMPLES);
}

float shadowCalculationDirectionalLight(int lightIndex, vec3 fragPos)
{
    vec4 frag_pos_lightspace = vert.frag_pos_lightspace[lightIndex];

    vec3 projCoords = frag_pos_lightspace.xyz / frag_pos_lightspace.w;
    if(projCoords.z > 1.0) {
		return 0.0;
	}

    projCoords = projCoords * 0.5 + 0.5;
    float bias = max(0.0005 * (1.0 - dot(vert.normalWorld, directionalLights[lightIndex].direction)), 0.00005);
    //float bias = max(0.05 * (1.0 - dot(vert.normalWorld, directionalLights[lightIndex].direction)), 0.005); 
    float currentDepth = projCoords.z - bias;

    float shadow = 0.0;
    vec2 texelSize = 1.0 / textureSize(directionalShadowMap[lightIndex], 0);
    for(int x = -PCF_COUNT; x <= PCF_COUNT; ++x)
    {
        for(int y = -PCF_COUNT; y <= PCF_COUNT; ++y)
        {
            float pcfDepth = texture(directionalShadowMap[lightIndex], (projCoords.xy + vec2(x, y) * texelSize)).r; 
            shadow += currentDepth - bias > pcfDepth ? 1.0 : 0.0;       
        }    
    }
    return shadow / PCF_TOTAL_SAMPLES;
}

float shadowCalculationSpotLight(int lightIndex, vec3 fragPos)
{
    vec4 frag_pos_lightspace = vert.frag_pos_lightspace_spot[lightIndex];

    vec3 projCoords = frag_pos_lightspace.xyz / frag_pos_lightspace.w;
    if(projCoords.z > 1.0) {
		return 0.0;
	}

    projCoords = projCoords * 0.5 + 0.5;
    float currentDepth = projCoords.z;

    float closestDepth = texture(spotShadowMap[lightIndex], projCoords.xy).r;
    float shadow = currentDepth > closestDepth ? 1.0 : 0.0; 
    return shadow;
}