#version 430

in vec2 outUVs;

layout(location = 0) out vec4 color;

const float PI_RCP = 0.318309886183790672;
const float PHASE_ISOTROPIC_FACTOR = 0.579577471545947667884;

const int POINT_NUM = 10;
const int SPOT_NUM = 10;
const int OMNI_SHADOW_MAPS = 3;
const int SPOT_SHADOW_MAPS = 1;
const int SPOT_SAMPLE_COUNT = 30;
const int POINT_SAMPLE_COUNT = 30;

// array sizes
uniform int actualPointSize;
uniform int actualSpotSize;

uniform sampler2D depthScene;
//uniform sampler2D scene;
uniform samplerCube omniShadowMap[OMNI_SHADOW_MAPS];
uniform sampler2D spotShadowMap[SPOT_SHADOW_MAPS];
uniform mat4 light_projection_matrices[SPOT_SHADOW_MAPS];
uniform mat4 light_view_matrices[SPOT_SHADOW_MAPS];

uniform mat4 projection_inv; 
uniform mat4 view_inv; 
uniform vec3 view_pos;

uniform float brightness; 

uniform float farPlane;

//noise calculation
uniform float time;

struct PointLight {
	vec3 color;
    float PADDING1;
	vec3 position;
    float PADDING2;
	vec3 attenuation;
    float fallOfValue;
    int shadowFBO;    
	int depthMap;
    int isActive; 
    int volumetric;
};

struct SpotLight {
    vec3 color;
	float cutOff;
	vec3 direction;
	float outerCutOff;
	vec3 attenuation;
    float PADDING1;
	vec3 position;
	float PADDING2;
	int shadowFBO;
	int depthMap;
	int isActive;
    int volumetric;
};

layout(std140) uniform pointBlock {
    PointLight pointLights[POINT_NUM];
};

layout(std140) uniform spotBlock {
    SpotLight spotLights[SPOT_NUM];
};

vec3 calc_world_pos_from_depth(float depth);
float tri(float x);
vec3 tri3(vec3 p);
float triNoise3d(vec3 p, float spd, float time);
vec3 calcVolumetricPointLight(vec3 frag_pos);
vec3 calcVolumetricSpotLight(vec3 frag_pos);

void main() 
{   
    vec3 colorTemp = vec3(0.0, 0.0, 0.0);    
    float depth = texture(depthScene, outUVs).r;
    vec3 frag_pos = calc_world_pos_from_depth(depth);         

    colorTemp += calcVolumetricPointLight(frag_pos);
    colorTemp += calcVolumetricSpotLight(frag_pos);

    color = vec4(colorTemp, depth);
}

//taken from Appendix (slides Revision Course)
vec3 calc_world_pos_from_depth(float depth)
{
    float z = depth * 2.0 - 1.0;
    vec4 clip_space_position = vec4(outUVs * 2.0 - 1.0, z, 1.0);
    vec4 view_space_position = projection_inv * clip_space_position;    
    view_space_position /= view_space_position.w;
    vec4 world_space_position = view_inv * view_space_position;
    return world_space_position.xyz;
}

//noise for volumetric lighting
float tri(float x)
{
    return abs(fract(x)-0.5);
}

vec3 tri3(vec3 p)
{
    return vec3( tri(p.z+tri(p.y*1.0)), tri(p.z+tri(p.x*1.0)), tri(p.y+tri(p.x*1.0)));
}    
    
float triNoise3d(vec3 p, float spd, float time){
    float z = 1.4; 
    float rz = 0.0; 
    vec3 bp = p;

    for(float i = 0.0; i <= 3.0; i++) 
    {
        vec3 dg = tri3(bp*2.0);
        p += (dg+time*spd);
        bp *= 1.8;
        z *= 1.5;
        p *= 1.2; 
        rz += (tri(p.z+tri(p.x+tri(p.y))))/z;
        bp += 0.14;
    }
    return rz;
}

float dither_pattern[16] = float[16] (
	0.0f, 0.5f, 0.125f, 0.625f,
    0.75f, 0.22f, 0.875f, 0.375f,
    0.1875f, 0.6875f, 0.0625f, 0.5625,
    0.9375f, 0.4375f, 0.8125f, 0.3125
);

vec3 calcVolumetricPointLight(vec3 frag_pos)
{
    vec3 colorTemp = vec3(0.0);

    for (int i=0; i < actualPointSize; i++)
    {   
        if(pointLights[i].isActive == 1 && pointLights[i].volumetric == 1)
        {
           float dither_value = dither_pattern[(int(gl_FragCoord.x) % 4)* 4 + (int(gl_FragCoord.y) % 4)];

            vec4 end_pos_worldspace = vec4(frag_pos, 1.0);
            vec4 start_pos_worldspace = vec4(view_pos, 1.0);
            vec4 delta_worldspace = normalize(end_pos_worldspace - start_pos_worldspace);

            float raymarch_distance_worldspace = length(end_pos_worldspace - start_pos_worldspace); 

            float step_size_worldspace = raymarch_distance_worldspace / POINT_SAMPLE_COUNT; 

            vec4 ray_position_worldspace = start_pos_worldspace + dither_value * step_size_worldspace;        

            float light_contribution = 0.0;
            float tau = 1.0;
            float power = 1.5;
            for (float l = raymarch_distance_worldspace; l > step_size_worldspace; l -= step_size_worldspace)
            {
                vec3 position_delta = ray_position_worldspace.xyz - pointLights[i].position;
                float d = length(position_delta);           
                float closest_depth = texture(omniShadowMap[i], position_delta).r;                
                closest_depth *= farPlane;

                float shadow_term = 1.0;
                if (d > closest_depth) {
                    shadow_term = 0.0;
                }           
                
                float fog = triNoise3d(ray_position_worldspace.xyz, 0.3, time)*2.0;
                
                float phi = power / (pointLights[i].attenuation.x + pointLights[i].attenuation.y * d + pointLights[i].attenuation.z * pow(d, 2)); // equals phi(d) in slides

                light_contribution += tau * shadow_term * fog * phi * step_size_worldspace * PHASE_ISOTROPIC_FACTOR * exp(-d/pointLights[i].fallOfValue);  // PHASE_ISOTROPIC_FACTOR = 1 / (4 * pi), exp(-d / fallof factor) is a fallof factor.
                
                ray_position_worldspace += step_size_worldspace * delta_worldspace;
                
            } 
            colorTemp += light_contribution * pointLights[i].color;
        }     
                
    }
    return colorTemp;
}

vec3 calcVolumetricSpotLight(vec3 frag_pos)
{
    vec3 colorTemp = vec3(0.0);

    for (int i=0; i < actualSpotSize; i++)
    {
        if(spotLights[i].isActive == 1 && spotLights[i].volumetric == 1)
        
        {
            float dither_value = dither_pattern[(int(gl_FragCoord.x) % 4) * 4 + (int(gl_FragCoord.y) % 4)];

            vec4 end_pos_worldspace  = vec4(view_pos, 1.0);
            vec4 start_pos_worldspace = vec4(frag_pos, 1.0);            
            vec4 delta_worldspace = normalize(end_pos_worldspace - start_pos_worldspace);
            
            vec4 end_pos_lightview = light_view_matrices[i] * end_pos_worldspace;
            vec4 start_pos_lightview = light_view_matrices[i] * start_pos_worldspace;
            vec4 delta_lightview = normalize(end_pos_lightview - start_pos_lightview);

            float raymarch_distance_lightview = length(end_pos_lightview - start_pos_lightview);
            float step_size_lightview = raymarch_distance_lightview / SPOT_SAMPLE_COUNT;

            float raymarch_distance_worldspace = length(end_pos_worldspace - start_pos_worldspace);
            float step_size_worldspace = raymarch_distance_worldspace / SPOT_SAMPLE_COUNT;

            vec4 ray_position_lightview = start_pos_lightview + dither_value * step_size_lightview * delta_lightview;
            vec4 ray_position_worldspace = start_pos_worldspace + dither_value * step_size_worldspace * delta_worldspace;

            float light_contribution = 0.0;
            float tau = 1.0; //0.02
            float epsilon = spotLights[i].cutOff - spotLights[i].outerCutOff;

            for (float l = raymarch_distance_worldspace; l > step_size_worldspace; l -= step_size_worldspace)
            {
                vec4 ray_position_lightspace = light_projection_matrices[i] * vec4(ray_position_lightview.xyz, 1);
                
                
                vec3 lightTmp = spotLights[i].position - ray_position_worldspace.xyz;
                float d = length(lightTmp);
                float d_rcp = 1.0/d;
                vec3 lightDir = normalize(lightTmp); // light direction                

                float theta = dot(lightDir, normalize(-spotLights[i].direction));
                if (theta > spotLights[i].outerCutOff) {
                    // We're inside the cone!
                    vec3 proj_coords = ray_position_lightspace.xyz / ray_position_lightspace.w;
                    proj_coords = proj_coords * 0.5 + 0.5;

                    float closest_depth = texture(spotShadowMap[i], proj_coords.xy).r;
                    float shadow_term = 1.0;
                    if (proj_coords.z > closest_depth) {
                        shadow_term = 0.0;
                    }
                    float cone_falloff = clamp((theta - spotLights[i].outerCutOff) / epsilon, 0.0, 1.0);

                    float strength = 1.0 / (spotLights[i].attenuation.x + spotLights[i].attenuation.y * d + spotLights[i].attenuation.z * pow(d, 2)); // equals phi(d) in slides
                    strength *= cone_falloff;
                    float fog = triNoise3d(ray_position_worldspace.xyz * 2.2 / 8, 0.1, time); //0.075
                    light_contribution += 7.0 * tau *fog* shadow_term * strength * step_size_worldspace * (1.0 - exp(-d*tau))*(1.0 - exp(-d*l));  // PHASE_ISOTROPIC_FACTOR = 1 / (4 * pi)
                    //light_contribution += fog * strength * tau * (shadow_term * (strength * 0.25 * PI_RCP) * d_rcp * d_rcp ) * exp(-d*tau)*exp(-l*tau) * step_size_worldspace;                    
                }
                ray_position_lightview += step_size_lightview * delta_lightview;
                ray_position_worldspace += step_size_worldspace * delta_worldspace;
            } 
            colorTemp += light_contribution * spotLights[i].color;
        } 
    }
    return colorTemp;
}